#!/usr/bin/env python

# DEMO PROGRAM FOR EX9000-MTCP SERIES MODULES
# DESIGNED & TESTED ON EX9050-MTCP
# LOGIC SUPPLY, 23 Jul 2015

import socket
import sys
import time

"""Usage notes"""
def print_usage():
    print "Usage: topsccc_demo_mtcp.py [--demo|--term-normal|--term-modbus|" + \
          "\t--demo          Run the demo program\r\n" + \
          "\t--term-normal   Normal prompt- enter and execute commands as documented\r\n" + \
          "\t--term-modbus   Modbus prompt- enter and execute commands as documented\r\n"
          
def send(command, quiet=False):    
    if type(command) is list:
        # preload packet with standard header values:
        # all packets begin with five bytes of zeros, the packet body length,
        # and have an module ID of 0x01, so we take care of it automatically
        pack = '\x00\x00\x00\x00\x00%s\x01' % chr(len(command) + 1)
        for i in packetize_CRC(command):
            pack = ''.join((pack, chr(i)))
        sock.sendto(pack, remote_addr)
        if not quiet:
            print 'Modbus packet: ',
            print modbus_str_to_hex(pack)
    else:
        sock.sendto(command + '\r\n', remote_addr)
        if not quiet:
            print 'Normal packet: ',
            print command

"""turn a modbus packet into a formatted list of bytes"""
def modbus_str_to_hex(packet):
    formatted_pack = ''
    for char in packet:
        formatted_pack = ''.join((formatted_pack, ' ', hex(ord(char))[2:].zfill(2)))
    return formatted_pack.upper()    

"""nice packet-making wrapper for modbus"""
def packetize_CRC(data):
    return data + modbusCRC(data)

"""modbus crc generator - crc is required for module modbus protocol"""
def modbusCRC(data):
    # CRC Generation
    # 1. Take 0xFFFF, XOR with the first byte in packet
    # 2. If LSB of result is 1, you will need to XOR a second value after performing step 3
    # 3. Shift result 1 place right.
    # 4. If the LSB at step 2 was 1, XOR the result with 0xA001
    # 5. Repeat steps 2-4 7 more times
    # 6. Repeat steps 2-5 for all bytes in packet
    # 7. The resulting CRC is 2 bytes. When sending the packet, the least significant BYTE
    #    is sent first, followed by the most significant BYTE.
    # result with the carry 0xA001
    crc = 0xFFFF # start off with 0xFFFF
    for byte in data: 
        crc = crc ^ (0x0000 | byte) # XOR each byte against existing CRC
        for i in range(8):
            carry = 0xA001 if crc & 0x0001 else 0x0000 # check LSB
            crc = (crc >> 1) ^ carry        #bit-shift right and XOR with 0xA001 if LSB eq'd 1
        crcl = crc % 256 # isolate LSB
        crch = (crc - crcl) >> 8 # shift MSB down 1 byte
    return [crcl, crch] # swap bytes and return

def end():
    sock.close()
    sys.exit(0)

def udp_socket():
    remote_addr = ("10.0.0.1", 1025)
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
    sock.settimeout(1)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.bind(("", 1025))
    return remote_addr, sock

def tcp_socket():
    remote_addr = ("10.0.0.1", 502)
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.settimeout(1)
    sock.connect(remote_addr)
    return remote_addr, sock

# MAIN FLOW
if "--help" in sys.argv:
    print_usage()
    sys.exit(0)

# Check for demo launch parameters
if len(sys.argv) == 1 or sys.argv[1] == '--demo':
    #ascii mode
    remote_addr, sock = udp_socket()
    send('$01M')
    result = sock.recvfrom(128)
    print "Found module %s at address %s" % (result[0][3:-1], result[1][0])
    # clear outputs
    send('$010000')
    time.sleep(0.4)
    # iterate through outputs
    for i in range(7):
        send('#0100%s' % hex(0x01 << i)[2:].upper().zfill(2))
        time.sleep(0.4)
    # fill outputs
    send('#01003F')
    time.sleep(0.4)
    # input status
    sock.recvfrom(1024) # clear socket buffer
    print '\nGetting input status: '
    send('$016')
    print 'Status: %s\n' % sock.recvfrom(15)[0][:-1]
    sock.close()
    #end ascii mode
    remote_addr, sock = tcp_socket()
    #clear outputs
    send([0x0F, 0x00, 0x10, 0x00, 0x06, 0x01, 0x00])
    time.sleep(0.4)
    # iterate through outputs
    for i in range(7):
        send([0x0F, 0x00, 0x10, 0x00, 0x06, 0x01, 0x01 << i])
        time.sleep(0.4)
    #fill outputs
    send([0x0F, 0x00, 0x10, 0x00, 0x06, 0x01, 0xFF])
    time.sleep(0.4)
    #input status
    sock.recvfrom(1024) # clear socket buffer
    print "\nGetting input status:"
    send([0x02, 0x00, 0x00, 0x00, 0x0C])
    print "Status: %s\n" % modbus_str_to_hex(sock.recvfrom(15)[0])
    end()
elif sys.argv[1] == '--term-normal':
    #terminal
    remote_addr, sock = udp_socket()
    print "\nEntering demo terminal mode.\nEnter commands as documented and press Enter to send.\nPress CTRL+C to exit.\n"
    try:
        while 1:
            command = ''.join((raw_input('Command: '), '\r\n',))
            send(command, quiet=True)
            print '  Reply: %s' % sock.recvfrom(100)[0][:-1]
    except KeyboardInterrupt:
        print "\nExiting demo terminal."
    end()
elif sys.argv[1] == '--term-modbus':
    remote_addr, sock = tcp_socket()
    print "\nEntering demo modbus terminal mode.\nEnter commands as space-separated hex bytes (0x00 0x01 ...) and \npress Enter to send.\nPress CRTL+C to exit.\n"
    while 1:
        command = []
        for byte in raw_input("Command: ").split(' '):
            command.append(int(byte[2:], 16))
        send(command, quiet=True)
        print "Reply:  ",
        #for char in sock.recvfrom(50)[0]:
        #    print "0x%s" % hex(ord(char))[2:].zfill(2),
        #print ""
        print modbus_str_to_hex(sock.recvfrom(50)[0])
    end()
